#include "RdmaCqProcessor.hpp"

RdmaCqProcessor::RdmaCqProcessor(ibv_context *verbs, int compQueueSize)
{
    CPPLog::LOG_ALWAYS("RdmaClientEndpointGroup : Creating CQ processor");
    //_qpEndpointMap = new std::unordered_map<uint32_t, RdmaClientEndpoint *>();
    _compChannel = ibv_create_comp_channel(verbs);
    if (_compChannel == NULL)
    {
        CPPLog::LOG_ERROR("CqProcessr : ibv_create_comp_channel failed");
        exit(1);
    }
    _completionQueue = ibv_create_cq(verbs, compQueueSize, NULL, _compChannel, 0);
    if (_completionQueue == NULL)
    {
        CPPLog::LOG_ERROR("CqProcessr : ibv_create_cq failed");
        exit(1);
    }
    int ret = ibv_req_notify_cq(_completionQueue, 0);
    if (ret)
    {
        CPPLog::LOG_ERROR("CqProcessr : ibv_req_notify_cq failed");
    }
}

struct ibv_cq *RdmaCqProcessor::getCq()
{
    return _completionQueue;
}

void RdmaCqProcessor::registerEp(uint64_t qp, RdmaClientEndpoint *ep)
{
    _qpEndpointMap.emplace(qp, ep);
}

void RdmaCqProcessor::deRegisterEp(uint64_t qp)
{
    _qpEndpointMap.erase(qp);
}

void RdmaCqProcessor::start()
{
    CPPLog::LOG_ALWAYS("CqProcessr : starting process CQ events");
    _compQueueThread = new std::thread(&RdmaCqProcessor::processCQEvents, this);
    pthread_setname_np(_compQueueThread->native_handle(), "compQueueThread");
}

void RdmaCqProcessor::processCQEvents()
{
    int ret = 0;
    struct ibv_cq *cq;
    void *context;
    const int nevent = 10;
    struct ibv_wc wc_array[nevent];
    while (_stop != true)
    {

        ret = ibv_get_cq_event(_compChannel, &cq, &context);
        if (ret == -1)
        {
            CPPLog::LOG_ERROR("CqProcessr : ibv_get_cq_event failed");
            close();
        }
        ibv_ack_cq_events(cq, 1);
        ret = ibv_req_notify_cq(_completionQueue, 0);
        if (ret)
        {
            CPPLog::LOG_ERROR("CqProcessr : ibv_req_notify_cq failed");
            close();
        }
        ret = ibv_poll_cq(cq, nevent, wc_array);
        if (ret < 0)
        {
            CPPLog::LOG_ERROR("CqProcessr : ibv_poll_cq failed");
            close();
        }
        if (ret == 0)
            continue;

        dispatchCqEvents(wc_array, ret);
    }
}

inline void RdmaCqProcessor::dispatchCqEvents(ibv_wc wc[], int size)
{
    for (int i = 0; i < size; i++)
    {
        if (wc[i].status != IBV_WC_SUCCESS)
        {
            std::cout << "RdmaCqProcessor : failed work completion : " << ibv_wc_status_str(wc[i].status) << " on qp " << wc[i].qp_num << std::endl;
            return;
        }
        auto it = _qpEndpointMap.find(wc[i].qp_num);
        if (it == _qpEndpointMap.end())
        {
            CPPLog::LOG_ALWAYS("RdmaCqProcessor : endpoint not registered for qp num");
            return;
        }
        switch (wc[i].opcode)
        {
        case IBV_WC_SEND:

            it->second->processSendComp(wc[i]);
            break;
        case IBV_WC_RECV:
            it->second->processRecvComp(wc[i]);
            break;
        case IBV_WC_RDMA_WRITE:
            std::cout << "rdma write completion\n";
            break;
        case IBV_WC_RDMA_READ:
            std::cout << "rdma read completion\n";
            break;
        default:
            std::cout << "RdmaCqProcessor : invalid opcode" << std::endl;
            break;
        }
    }
}

void RdmaCqProcessor::close()
{
    _stop = true;
    if(_compQueueThread != nullptr)
    {
        _compQueueThread->join();
        delete _compQueueThread;
    }
    if (_completionQueue != nullptr)
        ibv_destroy_cq(_completionQueue);
    if (_compChannel != nullptr)
        ibv_destroy_comp_channel(_compChannel);
    
}